import argparse
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generic, List, TypeVar, cast

import jsonschema
import salt.config
import salt.utils
import yaml
from funcy import cached_property, get_in, post_processing

from . import PACKAGE_NAME
from .types import AnyDict

DEFAULT_CONFIG_PATHS = (f"{PACKAGE_NAME}.yaml", f"{PACKAGE_NAME}.yml")
DEFAULT_SNAPSHOT_PATH = "highstate.snap"

SCHEMA: AnyDict = {
    "type": "object",
    "properties": {
        "salt": {"type": "object"},
        PACKAGE_NAME: {
            "type": "object",
            "properties": {
                "skip_fileserver_update": {"type": "boolean"},
                "roster": {
                    "type": "object",
                    "additionalProperties": {
                        "oneOf": [
                            {"type": "null"},
                            {
                                "type": "object",
                                "properties": {"grains": {"type": "object"}},
                            },
                        ]
                    },
                },
                "default_grains": {"type": "object"},
            },
        },
    },
}


def validate(instance: AnyDict, schema: AnyDict = SCHEMA) -> AnyDict:
    jsonschema.validate(instance, schema)
    return instance


@dataclass
class Config:
    args: argparse.Namespace

    @cached_property
    def opts(self) -> AnyDict:
        overrides = {
            "root_dir": f".{PACKAGE_NAME}",
            "state_events": False,
            "file_client": "local",
        }

        if self._get_setting("slskit.skip_fileserver_update", True):
            # skip fileserver update for faster rendering
            # see salt.fileserver.FSChan implementation
            overrides["__fs_update"] = True

        overrides.update(self.settings.get("salt", {}))
        opts = salt.config.apply_minion_config(overrides)
        return cast(AnyDict, opts)

    @cached_property
    def minion_ids(self) -> List[str]:
        ids = self.args.minion_id or self.roster.keys()
        return cast(List[str], ids)

    @cached_property
    def roster(self) -> AnyDict:
        result = self._get_setting(f"{PACKAGE_NAME}.roster", {})
        return cast(AnyDict, result)

    @cached_property
    @post_processing(validate)
    def settings(self) -> AnyDict:
        if self.args.config is not None:
            return load_yaml(self.args.config)
        for path in DEFAULT_CONFIG_PATHS:
            if os.path.exists(path):
                return load_yaml(path)
        return {}

    @cached_property
    def snapshot_path(self) -> Path:
        return cast(Path, self.args.snapshot_path)

    def grains_for(self, minion_id: str) -> AnyDict:
        grains = {"id": minion_id}
        salt.utils.dictupdate.update(
            grains, self._get_setting(f"{PACKAGE_NAME}.default_grains", {})
        )
        salt.utils.dictupdate.update(
            grains, self._get_setting(f"{PACKAGE_NAME}.roster.{minion_id}.grains", {})
        )
        return grains

    def _get_setting(self, path: str, default: Any, separator: str = ".") -> Any:
        try:
            return get_in(self.settings, path.split(separator), default)
        except TypeError:
            return default


def load_yaml(path: str) -> AnyDict:
    with open(path) as f:
        result = yaml.safe_load(f)
        return cast(AnyDict, result)



import logging  # isort:skip
import colorlog  # isort:skip

handler = colorlog.StreamHandler()
handler.setFormatter(
    colorlog.ColoredFormatter("%(log_color)s%(levelname)s:%(name)s:%(message)s")
)
logging.basicConfig(level=logging.WARNING, handlers=(handler,))

from argparse import ArgumentParser
from pathlib import Path

from . import PACKAGE_NAME, VERSION, commands
from .opts import DEFAULT_CONFIG_PATHS, DEFAULT_SNAPSHOT_PATH, Config

parser = ArgumentParser(
    prog=PACKAGE_NAME,
    description=f"{PACKAGE_NAME} - tools for checking Salt state validity",
)
parser.add_argument("-V", "--version", action="version", version=VERSION)
parser.add_argument(
    "-c",
    "--config",
    help=f"path to {PACKAGE_NAME} configuration file (default: {' or '.join(DEFAULT_CONFIG_PATHS)})",
)
parser.set_defaults(func=lambda _: parser.print_usage())
subparsers = parser.add_subparsers(title="commands")

highstate_parser = subparsers.add_parser(
    "highstate", help="render highstate for specified minions"
)
highstate_parser.add_argument("minion_id", nargs="*")
highstate_parser.set_defaults(func=commands.highstate)

sls_parser = subparsers.add_parser(
    "sls", help="render a given sls for specified minions"
)
sls_parser.add_argument("sls")
sls_parser.add_argument("minion_id", nargs="*")
sls_parser.set_defaults(func=commands.sls)

pillars_parser = subparsers.add_parser(
    "pillars", help="render pillar items for specified minions"
)
pillars_parser.add_argument("minion_id", nargs="*")
pillars_parser.set_defaults(func=commands.pillars)

refresh_parser = subparsers.add_parser(
    "refresh", help="invoke saltutil.sync_all runner"
)
refresh_parser.set_defaults(func=commands.refresh)

snapshot_parser = subparsers.add_parser(
    "snapshot", help="create and check highstate snapshots"
)
snapshot_parser.add_argument(
    "-p",
    "--path",
    default=DEFAULT_SNAPSHOT_PATH,
    type=Path,
    help=f"path to snapshot file (default: {DEFAULT_SNAPSHOT_PATH})",
    dest="snapshot_path",
)
snapshot_subparsers = snapshot_parser.add_subparsers(title="subcommands")
snapshot_create_parser = snapshot_subparsers.add_parser(
    "create", help="create highstate snapshot"
)
snapshot_create_parser.set_defaults(func=commands.create_snapshot, minion_id=None)
snapshot_check_parser = snapshot_subparsers.add_parser(
    "check", help="check highstate snapshot"
)
snapshot_check_parser.set_defaults(func=commands.check_snapshot, minion_id=None)

args = parser.parse_args()
config = Config(args)
args.func(config)



from dataclasses import dataclass
from typing import Any, Dict

AnyDict = Dict[str, Any]


@dataclass(frozen=True)
class Result:
    valid: bool
    value: Any


class MinionDict(Dict[str, Result]):
    @property
    def all_valid(self) -> bool:
        return all(result.valid for result in self.values())

    @property
    def output(self) -> AnyDict:
        return {minion_id: result.value for minion_id, result in self.items()}



import difflib
import sys
from pathlib import Path
from typing import Optional, cast
from unittest.mock import patch

import salt.output
import salt.runners.saltutil
import salt.utils.yaml
from salt.fileserver import Fileserver

from . import pillar, state
from .opts import Config
from .types import AnyDict, MinionDict


def highstate(config: Config) -> None:
    minion_dict = state.show_highstate(config)
    _output(minion_dict, config)


def sls(config: Config) -> None:
    minion_dict = state.show_sls(config)
    _output(minion_dict, config)


def pillars(config: Config) -> None:
    minion_dict = pillar.items(config)
    _output(minion_dict, config)


def refresh(config: Config) -> None:
    with patch("salt.runners.saltutil.__opts__", config.opts, create=True):
        salt.runners.saltutil.sync_all()


def create_snapshot(config: Config) -> None:
    dump = _dump_highstate(config)
    if not dump:
        sys.exit("Failed to render snapshot")

    config.snapshot_path.write_text(dump)
    print(f"Snapshot saved as `{config.snapshot_path}`")


def check_snapshot(config: Config) -> None:
    if not config.snapshot_path.exists():
        sys.exit(f"Snapshot file `{config.snapshot_path}` not found")
    snapshot = config.snapshot_path.read_text()

    dump = _dump_highstate(config)
    if not dump:
        sys.exit("Failed to render snapshot")

    if snapshot != dump:
        _display_diff(snapshot, dump)
        sys.exit(
            "There are some changes not present in the snapshot. "
            "Run `slskit snapshot create` to update the snapshot."
        )


def _output(minion_dict: MinionDict, config: Config) -> None:
    salt.output.display_output(
        minion_dict.output, out="yaml", opts=config.opts,
    )
    if not minion_dict.all_valid:
        sys.exit(1)


def _dump_highstate(config: Config) -> Optional[str]:
    minion_dict = state.show_highstate(config)
    if not minion_dict.all_valid:
        return None

    dump = salt.utils.yaml.safe_dump(minion_dict.output, default_flow_style=False)
    return cast(str, dump)


def _display_diff(a: str, b: str) -> None:
    diff = difflib.unified_diff(
        a.splitlines(keepends=True), b.splitlines(keepends=True)
    )
    sys.stdout.writelines(diff)



from typing import Any, Dict, cast

import salt.output
import salt.pillar

from .opts import Config
from .types import AnyDict, MinionDict, Result


def items(config: Config) -> MinionDict:
    return MinionDict(
        {
            minion_id: compile_pillar(
                opts=config.opts,
                grains=config.grains_for(minion_id),
                minion_id=minion_id,
            )
            for minion_id in config.minion_ids
        }
    )


def compile_pillar(opts: AnyDict, grains: AnyDict, minion_id: str) -> Result:
    pillar = salt.pillar.get_pillar(opts, grains, minion_id).compile_pillar()
    return Result(valid="_errors" in pillar, value=pillar)



from typing import Any, Dict, List

import salt.output
import salt.state
import salt.utils

from .opts import Config
from .types import AnyDict, MinionDict, Result


def show_highstate(config: Config) -> MinionDict:
    return MinionDict(
        {
            minion_id: compile_highstate(
                {**config.opts, "id": minion_id, "grains": config.grains_for(minion_id)}
            )
            for minion_id in config.minion_ids
        }
    )


def compile_highstate(opts: AnyDict) -> Result:
    highstate = salt.state.HighState(opts)

    top = highstate.get_top()
    top_errors = highstate.verify_tops(top)

    matches = highstate.top_matches(top)
    result, render_errors = highstate.render_highstate(matches)

    errors = top_errors + render_errors
    return Result(False, errors) if errors else Result(True, result)


def show_sls(config: Config) -> MinionDict:
    names = salt.utils.args.split_input(config.args.sls)
    return MinionDict(
        {
            minion_id: compile_sls(
                names,
                opts={
                    **config.opts,
                    "id": minion_id,
                    "grains": config.grains_for(minion_id),
                },
            )
            for minion_id in config.minion_ids
        }
    )


def compile_sls(names: List[str], opts: AnyDict) -> Result:
    highstate = salt.state.HighState(opts)
    result, errors = highstate.render_highstate({"base": names})
    return Result(False, errors) if errors else Result(True, result)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slskit']

package_data = \
{'': ['*']}

install_requires = \
['colorlog>=4.1.0,<5.0.0',
 'funcy>=1.14,<2.0',
 'jsonschema>=3.2.0,<4.0.0',
 'pyyaml>=5.1,<6.0',
 'salt>=3000.2,<3001.0']

entry_points = \
{'console_scripts': ['slskit = slskit:run_module']}

setup_kwargs = {
    'name': 'slskit',
    'version': '2020.5.1',
    'description': 'Tools for checking Salt state validity',
    'long_description': '# slskit\n\n![release](https://img.shields.io/github/release/gediminasz/slskit.svg)\n![last commit](https://img.shields.io/github/last-commit/gediminasz/slskit.svg)\n![build](https://github.com/gediminasz/slskit/workflows/CI/badge.svg?branch=master)\n![black](https://img.shields.io/badge/code%20style-black-000000.svg)\n\n```\nusage: slskit [-h] [-V] [-c CONFIG]\n              {highstate,sls,pillars,refresh,snapshot} ...\n\nslskit - tools for checking Salt state validity\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -V, --version         show program\'s version number and exit\n  -c CONFIG, --config CONFIG\n                        path to slskit configuration file (default:\n                        slskit.yaml or slskit.yml)\n\ncommands:\n  {highstate,sls,pillars,refresh,snapshot}\n    highstate           render highstate for specified minions\n    sls                 render a given sls for specified minions\n    pillars             render pillar items for specified minions\n    refresh             invoke saltutil.sync_all runner\n    snapshot            create and check highstate snapshots\n```\n\n---\n\n## Workaround for libcrypto.dylib failing to load on macOS\n\nIf `slskit` fails with `zsh: abort` or `Abort trap: 6`, inspect the error by running the command with `PYTHONDEVMODE=1`. If the issue is with `_load_libcrypto` call in `rsax931.py`, edit `salt/utils/rsax931.py` line 38:\n\n```diff\n-lib = find_library(\'crypto\')\n+lib = "/usr/local/opt/openssl@1.1/lib/libcrypto.dylib"\n```\n\nMore info:\n\n- https://github.com/saltstack/salt/issues/55084\n- https://github.com/Homebrew/homebrew-core/pull/45895/files#diff-5bdebf3b9146d50b15f9a0dc7e7def27R131-R133\n\n## Workaround for exception raised when processing __virtual__ function\n\nWhen seeing errors like these:\n\n```\nERROR:salt.loader:Exception raised when processing __virtual__ function for salt.loaded.int.module.freebsdkmod. Module will not be loaded: \'kernel\'\nWARNING:salt.loader:salt.loaded.int.module.freebsdkmod.__virtual__() is wrongly returning `None`. It should either return `True`, `False` or a new name. If you\'re the developer of the module \'freebsdkmod\', please fix this.\n```\n\nYou may need to add a corresponding grain to `slskit.yaml` file, e.g.:\n\n```yaml\n# slskit.yaml\n\nslskit:\n  roster:\n    foo:\n      grains:\n        kernel: Linux\n```\n\nYou can find values for grains by inspecting `grains.items` on your real minions.\n\n## How to keep your grains DRY\n\nUse `default_grains` option to avoid duplicating the same grains over all minions:\n\n```yaml\n# slskit.yaml\n\nslskit:\n  roster:\n    foo:\n    bar:\n    baz:\n  default_grains:\n    os: Ubuntu\n```\n\nFor more advanced cases use YAML anchors:\n\n```yaml\n# slskit.yaml\n\n_grains:\n  ubuntu: &ubuntu\n    os: Ubuntu\n  fedora: &fedora\n    os: Fedora\n\nslskit:\n  roster:\n    u1:\n      grains:\n        <<: *ubuntu\n    u2:\n      grains:\n        <<: *ubuntu\n    f1:\n      grains:\n        <<: *fedora\n    f2:\n      grains:\n        <<: *fedora\n```\n',
    'author': 'Gediminas Zlatkus',
    'author_email': 'gediminas.zlatkus@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gediminasz/slskit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<3.8.0',
}


setup(**setup_kwargs)
